%
%   PCC2 Standard Library
%
%   This file defines commands and keys used internally by PCC2,
%   and a number of user-callable routines described in the
%   manual. Do not modify unless you are absolutely 100% positively
%   sure what you are doing.
%
%   This file will be overwritten by PCC2 upgrades. It will most
%   likely only work with the PCC2 version it was intended for.
%
%   This version of core.q is for PCC2, and is incompatible with
%   PCC 1.x - and vice versa.
%

%Option Optimize(2), Lexical

Dim Shared CC$LibraryVersion = '1.99.19'

%%% Console-Mode Replacements for GUI routines %%%%%%%%%%%%%%%%%%%%%%

If Not System.GUI Then

  %%% Utilities

  % Return the index of the first word in list which starts with the given word.
  Function CC$MatchWord (list, word)
    Local i:=1, e
    Do While list<>''
      e := First(' ', list)
      If Mid(e,1,Len(word)) = word Then Return i
      i := i + 1
      list := Trim(Rest(' ', list))
    Loop
  EndFunction

  % Count number of words in list
  Function CC$CountWords (list)
    Local i:=0
    Do While list<>''
      i:=i+1
      list:=Trim(Rest(' ', list))
    Loop
    Return i
  EndFunction


  %%% Actual User Functions

  % MessageBox generates a console message.
  Sub MessageBox (text, Optional heading)
    % FIXME: PCC 1.x replaces \r by \n in text.
    If IsEmpty(heading) Then heading := Translate("Message")
    Print heading, ": ", text
  EndSub

  % UI.Update, UI.PopupConsole has no effect in console mode.
  Sub UI.Update (Optional flag)
  EndSub

  Sub UI.PopupConsole
  EndSub

  % Commands that call a dialog generate an error in console mode.
  Sub UI.ChooseObject (screen)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.Search (query, Optional flags)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.ListShips (X, Y, Optional flags, ok, Heading)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.ListFleets (X, Y, Optional flags, ok, Heading)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.GotoScreen (sid, Optional oid)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.GotoChart (X, Y)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.SelectionManager
    Abort "Not in graphics mode"
  EndSub

  Sub UI.PlanetInfo (pid)
    Abort "Not in graphics mode"
  EndSub

  Sub UI.KeymapInfo
    Abort "Not in graphics mode"
  EndSub

  Sub UI.Help (page)
    Abort "Not in graphics mode"
  EndSub

  % UI.InputNumber generates an error in console mode (like it does in PCC 1.x,
  % although we can probably do better).
  Sub UI.InputNumber (prompt, Optional lo, hi, cur, help)
    Abort "Not in graphics mode"
  EndSub

  % UI.Message generates a console message and a prompt
  Sub UI.Message (text, Optional heading, buttons)
    MessageBox text, Heading
    If IsEmpty(buttons) Then buttons := Translate("OK")
    Do
      UI.Input buttons
      If IsEmpty(UI.Result) Then
        % This means the last choice
        UI.Result := Z(CC$CountWords(buttons))
        Break
      Else If UI.Result = "" Then
        % This means the first choice
        UI.Result := 1
        Break
      Else
        % Does this match any button?
        UI.Result := CC$MatchWord(buttons, UI.Result)
        If Not IsEmpty(UI.Result) Then Break
      EndIf
    Loop
  EndSub

  % UI.OverlayMessage generates a console message
  Sub UI.OverlayMessage (text)
    Print Translate("Message") # ": " # text
  EndSub
EndIf


%%% User-callable Subroutines %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Sub WaitOneTurn ()
  Local t = Turn
  Do While t = Turn
    Stop
  Loop
EndSub

% Usable only in ship context: set ship waypoint, and suspend until
% the ship arrives there. Special case for gravity wells. We terminate
% when the waypoint is zero: when moving across a map border, we never
% reach the specified x,y.
Sub MoveTo (x, y)
  Local pid = PlanetAt(x, y, 1)
  With Lock("s" & Id & ".waypoint")
    SetWaypoint x, y
    Do Until (Waypoint.DX=0 And Waypoint.DY=0) Or ((pid<>0) And (Orbit$=pid))
      Stop
    Loop
  EndWith
EndSub

Sub MoveTowards (x, y)
  With Lock("s" & Id & ".waypoint")
    SetWaypoint x, y
    WaitOneTurn
  EndWith
EndSub

% Enqueue a ship of type h (1..105), engine e (1..9) with
% bc beams of type bt, and tc torpedo launchers of type tt.
% Suspend until this order was enqueued successfully.
Sub EnqueueShip (h, e, Optional bt, bc, tt, tc)
  % wait until pending build order performed
  Do While Build.YesNo
    Stop
  Loop
  % wait until build order successfully delivered
  Do
    Try
      BuildShip h, e, bt, bc, tt, tc
      Break
    EndTry
    Stop
  Loop
EndSub

% Sub CargoUploadWait (cargo, Optional flags)
% Sub CargoTransferWait (sid, cargo, Optional flags)

% Unload all ships at this location. Can be called from ship or planet.
Sub CargoUnloadAllShips
  Local thisX = Loc.X, thisY = Loc.Y
  Local Cargo.Remainder
  If Not PlanetAt(thisX, thisY) Then
    Abort "No planet at this location"
  Else
    ForEach Ship Do
      If Loc.X=thisX And Loc.Y=thisY And Played Then
        CargoUnload "10000tdmcs$", "n"
      EndIf
    Next
  EndIf
EndSub

% Building with waiting
Sub BuildMinesWait (amount)
  Local Build.Remainder = amount
  Do
    BuildMines Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

Sub BuildFactoriesWait (amount)
  Local Build.Remainder = amount
  Do
    BuildFactories Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

Sub BuildDefenseWait (amount)
  Local Build.Remainder = amount
  Do
    BuildDefense Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

Sub BuildBaseDefenseWait (amount)
  Local Build.Remainder = amount
  Do
    BuildBaseDefense Build.Remainder, "n"
    If Not Build.Remainder Then Return
    Stop
  Loop
EndSub

% Change this fleet's name
Sub SetFleetName (name)
  If Not IsEmpty(name) Then
    If Fleet$=0 Then Abort "Not in a fleet"
    Ship(Fleet$).Fleet.Name := name
  EndIf
EndSub

% Change leader of fleet fid to sid
Sub ChangeFleetLeader (fid, sid)
  If fid=sid Then Return
  % start the new fleet
  With Ship(sid) Do
    SetFleet sid
    Fleet.Name := Ship(fid).Fleet.Name
  EndWith
  % First move all members to the new fleet, then move the old leader
  ForEach Ship Do If (Id<>fid) And (Fleet$=fid) Then SetFleet sid
  With Ship(fid) Do SetFleet sid
EndSub

% Set selection layer. Convenience routine to replace direct assignment to Selection.Layer.
% Accepts selection layer number (0-7) or name ('A'..'H')
Sub SelectionLayer (l)
  If IsEmpty(l) Then Return
  If IsString(l) Then l:=InStr("abcdefgh", l)-1
  Selection.Layer := l
EndSub

% Sub FixShipWait (Optional id)
% Sub RecycleShipWait (Optional id)

% Notify message. Add message and stop the process until the user confirms.
% This is implemented totally different from PCC 1.x.
Sub Notify (s, Optional t)
  If Not IsEmpty(t) Then
    Do While Turn < t
      Stop
    Loop
  EndIf
  Do
    CC$Notify s, True
    Stop
  Loop While Not CC$NotifyConfirmed()
EndSub

% Add notify message without associating it with a process.
Sub AddNotify (s)
  CC$Notify s, False
EndSub

% Internal: check before 'Restart' command. This makes sure we do not
% enter infinite loops. 'Restart' is internally coded as a call to
% 'CC$AutoReCheck' before the loop.
Sub CC$AutoReCheck
  % Create process-local variable upon first entry into this routine
  Dim Static CC$AutoReState
  % Error when we're getting here again with a turn number we've already seen
  If CC$AutoReState = Turn
    Notify Translate("This Auto Task performed a \"Restart\" loop without intervening delays. You should probably use \"WaitOneTurn\".")
    CC$AutoReState := 0
  Else
    CC$AutoReState := Turn
  EndIf
EndSub

% Internal: execute auto task command. An auto task command 'foo' is
% actually coded as 'CC$AutoExec "foo"'. This implements error handling.
Sub CC$AutoExec(cmd)
  Do
    Try
      Eval cmd
      Return
    Else
      Notify Translate("Error during execution of auto task: ") + System.Err
    EndTry
  Loop
EndSub

% Sub AutoLoad (file, name, Optional args)

Function FindShipCloningAt (pid)
  If Cfg("AllowShipCloning")
    % FindShip returns the first ship, which is just what we need
    Return FindShip(Orbit$=pid And StrCase(FCode="cln"))
  Else
    Return Z(0)
  EndIf
EndFunction

% Get name of a ship, together with its Id. In PCC 1.x, this is what the
% "ShipName" builtin function does. In PCC2, "ShipName" returns the plain
% name which I believe is more useful, and consistent with "PlanetName".
% Code that used "ShipName" can now use "ShipNameAndId".
Function ShipNameAndId (sid)
  Local N = ShipName(sid)
  If N
    Return N & " (#" & sid & ")"
  Else
    Return Format(Translate("Ship %d"), sid)
  EndIf
EndFunction

%%% GUI Subroutines %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

If System.GUI Then
%% When the GUI is not active, don't define all these to save memory

Dim Shared CC$BuildH, CC$BuildE, CC$BuildBT, CC$BuildBC, CC$BuildTT, CC$BuildTC

% Ctrl-P on starbase/ship build screen
Sub CC$PickupBuild
  If Build.YesNo Then
    CC$BuildH  := Build.Hull$
    CC$BuildE  := Build.Engine$
    CC$BuildBT := Build.Beam$
    CC$BuildBC := Build.Beam.Count
    CC$BuildTT := Build.Torp$
    CC$BuildTC := Build.Torp.Count
    Messagebox Format(Translate("Ship specification for a %s saved. Use [Ctrl-B] on other starbases to build this ship there."), Build.Short)
  EndIf
EndSub

% Ctrl-B on starbase screen
Sub CC$PerformBuild
  Local System.Err
  If IsEmpty(CC$BuildH) Or Not Played Then Return
  If Build.YesNo Then
    UI.Message Format(Translate("This starbase is already building a %s. Override this with the saved build order for a %s?"), Build.Short, Hull(CC$BuildH).Name.Short), Translate("Ship Building"), Translate("Yes No")
    If UI.Result<>1 Then Return
  EndIf
  Try
    BuildShip CC$BuildH, CC$BuildE, CC$BuildBT, CC$BuildBC, CC$BuildTT, CC$BuildTC
  Else
    Messagebox Format(Translate("Unable to build a %s: %s"), Hull(CC$BuildH).Name.Short, System.Err)
  EndTry
EndSub

% Ctrl-S on ship build screen (not yet documented because it's weird).
% This is still the inelegant PCC 1.x version; we could do better by
% using a user-defined function.
Sub CC$SearchBuild
  % Value of tech level t = 50*(t-1)*t
  If Build.YesNo Then
    Local mstr := Hull(Build.Hull$).Cost.Str
    Local tstr := "Max(0,Tech.Hull*(Tech.Hull-1)-" & (Hull(Build.Hull$).Tech.Hull*(Hull(Build.Hull$).Tech.Hull-1)) & ")"
    mstr := CAdd(mstr, CMul(Engine(Build.Engine$).Cost.Str, Hull(Build.Hull$).Engine.Count))
    tstr := tstr & "+Max(0,Tech.Engine*(Tech.Engine-1)-" & (Engine(Build.Engine$).Tech*(Engine(Build.Engine$).Tech-1)) & ")"
    If Build.Beam.Count Then
      mstr := CAdd(mstr, CMul(Beam(Build.Beam$).Cost.Str, Build.Beam.Count))
      tstr := tstr & "+Max(0,Tech.Beam*(Tech.Beam-1)-" & (Beam(Build.Beam$).Tech*(Beam(Build.Beam$).Tech-1)) & ")"
    EndIf
    If Build.Torp.Count Then
      mstr := CAdd(mstr, CMul(Torpedo(Build.Torp$).Cost.Str, Build.Torp.Count))
      tstr := tstr & "+Max(0,Tech.Torpedo*(Tech.Torpedo-1)-" & (Torpedo(Build.Torp$).Tech*(Torpedo(Build.Torp$).Tech-1)) & ")"
    EndIf
    UI.Search "Not Build.YesNo And CCompare(Mined.Str&(Supplies+Money)&'$','" & mstr & "'&(" & tstr & ")*50&'$')", "2b"
  EndIf
EndSub

% '.' everywhere
Sub CC$ToggleSelection
  Mark Not Marked
EndSub

% F9 on ship
Sub CC$SetShipComment
  UI.Input Translate("Enter new comment for this ship:"), Translate("Edit Comment"), 255, "gm30", Comment
  SetComment UI.Result
EndSub

% F9 on planet
Sub CC$SetPlanetComment
  UI.Input Translate("Enter new comment for this planet:"), Translate("Edit Comment"), 255, "gm30", Comment
  SetComment UI.Result
EndSub

% Shift-F9 on control screens
Sub CC$SetPlanetCommentAt(X,Y)
  Local pid = PlanetAt(X,Y,1)
  If pid Then With Planet(pid) Do CC$SetPlanetComment
EndSub

% N in ship screen
Sub CC$ShipName
  If Not Played Then Return
  UI.Input Translate("Enter new name:"), Format(Translate("Rename Starship #%d"), Id), 20, "gm20", Name
  SetName UI.Result
EndSub

% Sub CC$FleetName
% Sub CC$ChangeLeader

% C-z
Sub CC$SetWaypoint (x, y)
  Try With Lock("s" & Id & ".waypoint") Do SetWaypoint x, y
EndSub

% Alt-Left/Alt-Right
Sub CC$SwitchSelection (delta)
  Selection.Layer := (Selection.Layer + 8 + delta) Mod 8
  UI.OverlayMessage Format(Translate("Selection %c"), Chr(Selection.Layer + 65))
EndSub

% Shift/Ctrl-F2/F3/F8
Sub CC$PlanetAt (typ, X, Y)
  Local pid := PlanetAt(X, Y, 1)
  If pid Then Try UI.GotoScreen typ, pid
EndSub

% Shift/Ctrl-F4
Sub CC$GotoChart (X, Y)
  If IsEmpty(X) Or IsEmpty(Y) Then
    UI.GotoScreen 4
  Else
    UI.GotoChart X,Y
  EndIf
EndSub

% L, Shift-L, Shift-F1, etc.
Sub CC$ListShips (X, Y, flags)
  Local UI.Result
  UI.ListShips X, Y, flags
  If UI.Result
    UI.GotoScreen 1, UI.Result
  EndIf
EndSub

% X on ship screen
Sub CC$ShipXchg
  Local myX = UI.X, myY = UI.Y, myId = Id
  If FindShip(Loc.X=myX And Loc.Y=myY And Id<>myId And Played)
    % There is one of our ships other than this one at this location
    CC$ListShips myX, myY, 'e' & myId
  Else
    % No ship, so check for a planet
    Local pid = PlanetAt(myX, myY, 1)
    If pid And Planet(pid).Played Then UI.GotoScreen 2, pid
  EndIf
EndSub

% X on planet screen
Sub CC$PlanetXchg
  Local pid = PlanetAt(UI.X, UI.Y, 1)
  If pid And pid<>Id And Planet(pid).Played Then
    UI.GotoScreen 2, pid
  Else
    If FindShip(Loc.X=UI.X And Loc.Y=UI.Y And Played) Then CC$ListShips UI.X, UI.Y, 'e'
  EndIf
EndSub

% X on base screen
Sub CC$BaseXchg
  Local pid = PlanetAt(UI.X, UI.Y, 1)
  If pid And pid<>Id And Planet(pid).Played Then
    If Planet(pid).Base.YesNo Then
      UI.GotoScreen 3, pid
    Else
      UI.GotoScreen 2, pid
    EndIf
  Else
    If FindShip(Loc.X=UI.X And Loc.Y=UI.Y And Played) Then CC$ListShips UI.X, UI.Y, 'e'
  EndIf
EndSub

% F1/F2/F3
Sub CC$GotoScreen (Screen)
  Local UI.Result
  UI.ChooseObject Screen
  If Not IsEmpty(UI.Result) Then UI.GotoScreen Screen, UI.Result
EndSub

% F1/F2/F3 on starchart
Sub CC$GotoObject (Screen, Id)
  If Id Then Try UI.GotoScreen Screen, Id
EndSub

% F1 on history screen
Sub CC$ShipScreenFromHistory
  If Played Then
    UI.GotoScreen 1, Id
  Else
    CC$GotoScreen 1
  EndIf
EndSub

% F5
Sub CC$PlanetInfo (X, Y)
  Local pid = PlanetAt(X,Y,True)
  If pid Then UI.PlanetInfo pid
EndSub

% F8
Sub CC$BuildOrGotoBase
  If Base.YesNo
    Try UI.GotoScreen 3, Id
  Else
    Try CC$BuildBase
  EndIf
EndSub

% ESC on race screen
Sub CC$ExitRace
  Local UI.Result
  UI.Message Translate("Do you want to exit this game?"), Translate("Exit Game"), Translate("Yes No")
  If UI.Result=1 Then System.ExitRace
EndSub

% Scanner movement
Sub CC$ScannerMoveX (delta)
  If UI.Prefix Then delta := If(delta < 0, -UI.Prefix, UI.Prefix)
  UI.X := UI.X + delta
EndSub

Sub CC$ScannerMoveY (delta)
  If UI.Prefix Then delta := If(delta < 0, -UI.Prefix, UI.Prefix)
  UI.Y := UI.Y + delta
EndSub

% B on starchart in planet view
Sub CC$BaseLock
  If Base.YesNo Then
    Chart.SetView 'BaseLock'
  Else
    Try CC$BuildBase
  EndIf
EndSub

% P on starchart in starbase view
Sub CC$PlanetLock
  Chart.SetView 'PlanetLock'
EndSub

% ESC on starchart. Since this hides the view, it will also turn off the keymap,
% so the next ESC will exit the map.
Sub CC$HidePanel
  Chart.SetView ''
EndSub

% C-p/C-s on ship screen. Only review the transfer if it is present.
Sub CC$ReviewShipTransfer (kind, Id)
  If Id Then CC$TransferShip kind, Id
EndSub

% c on ship/history screen
Sub CC$ShipCargo
  If Played Then
    CC$TransferShip 0,0
  Else
    CC$CargoHistory
  EndIf
EndSub

% u on ship screen
Sub CC$ShipUnload
  If Played Then
    If Orbit$ Then
      CC$TransferUnload
    Else
      MessageBox Translate("We are not orbiting a planet, sir. If you want to jettison cargo, use the 'c' command instead."), Translate("Cargo Transfer")
    EndIf
  EndIf
EndSub

% Ship operation
Sub CC$WithShipWaypoint (cmd, opt, Optional overrideAT)
  Local UI.Result
  Local sid = Id
  Local s, T
  If IsEmpty(overrideAT) Then overrideAT:=False

  % Check waypoint lock
  s := "s" & sid & ".waypoint"
  If Not overrideAT And GetLockInfo(s, 0) Then
    If GetLockInfo(s, 2) Then
      UI.Message Format(Translate("%s (%s)\nProceed anyway?"), GetLockInfo(s, 2), GetLockInfo(s, 1)), Translate("Waypoint"), Translate("Yes No")
    Else
      UI.Message Format(Translate("This function is under control of \"%s\".\nProceed anyway?"), GetLockInfo(s, 1)), Translate("Waypoint"), Translate("Yes No")
    EndIf
    If UI.Result <> 1 Then Return
  EndIf

  % Check fleet
  If Fleet$<>0 And Fleet$<>sid Then
    UI.Message Translate("This ship is member of a fleet. Do you want to set the waypoint of the whole fleet, or do you want to leave the fleet?"), Translate("Waypoint"), Translate("Set Leave Cancel")
    If UI.Result=3 Then Return
    If UI.Result=1 Then
      % Set
      sid:=Fleet$
    Else
      % Leave fleet
      Fleet$:=0
    EndIf
  EndIf

  % All tests passed
  With Ship(sid) Do Eval cmd

  % Set optimum speed
  If opt Then
    If Ship(sid).Speed<>"Hyperdrive" And Ship(sid).Waypoint.Dist>0 Then
      % Find out how long it takes to get there at this speed
      s := Global.Engine(Ship(sid).Engine$).Speed$
      Ship(sid).Speed$ := s
      T := Ship(sid).Move.Eta
      If T < 30 Then % FIXME: hardcoded value
        % Find whether we can reach the target with a slower speed
        Do While s > 1
          s := s-1
          Ship(sid).Speed$ := s
          If Ship(sid).Move.Eta > T Then
            % It's slower, so undo and stop
            Ship(sid).Speed$ := s+1
            Break
          EndIf
        Loop
      EndIf
    EndIf
  EndIf
EndSub

Sub CC$WithFleetWaypoint (cmd, opt)
  With Ship(Fleet$) Do CC$WithShipWaypoint cmd, opt
EndSub

% 'j' on fleet screen
Sub CC$FleetJoin
  Local UI.Result
  Local fid = Fleet$
  UI.ListFleets Ship(fid).Loc.X, Ship(fid).Loc.Y, "a" & fid, Translate("OK"), Translate("Join Fleet")
  If Not IsEmpty(UI.Result) Then
    % Move members, then leader
    ForEach Ship Do If Id<>fid And Fleet$=fid Then Fleet$:=UI.Result
    Ship(fid).Fleet$:=UI.Result
  EndIf
EndSub

% 'n' on fleet screen
Sub CC$FleetName
  Local UI.Result
  Local fid = Fleet$
  UI.Input Translate("Enter new name:"), Format(Translate("Rename Fleet #%d"), fid), 255, "gm20", Ship(fid).Fleet.Name
  SetFleetName UI.Result
EndSub

% 'p' on fleet screen
Sub CC$FleetSplit
  Local UI.Result
  Local oldfid = Fleet$                       % old fleet Id
  Local newmark = Not Ship(oldfid).Marked     % selection status for new fleet
  Local newcount = 0                          % number of ships in new fleet
  Local newfid = If (Marked=newmark, Id, 0)   % new leader is this ship if it has the right status

  % Count ships
  ForEach Ship Do
    If Fleet$=oldfid And Marked=newmark Then
      newcount:=newcount+1
      If newfid=0 Then newfid:=Id
    EndIf
  Next
  If newcount=0 Then
    MessageBox Translate("Please select the ships which should go into one fleet, and unselect those which should go into the other."), Translate("Split Fleet")
    Return
  EndIf

  % Confirm
  UI.Message Format(Translate("Form a new fleet with %d ship%!1{s%}, leader is %s?"), newcount, Ship(newfid).Name), Translate("Split Fleet"), Translate("Yes No")
  If UI.Result<>1 Then Return

  % Do it
  Ship(newfid).Fleet$ := newfid
  ForEach Ship Do If Fleet$=oldfid And Marked=newmark Then Fleet$:=newfid

  % Bring user there
  If Fleet$=newfid Then Iterator(10).CurrentIndex:=newfid
EndSub

% C-ins on fleet screen
Sub CC$FleetAddToSim
  Local fid = Fleet$
  ForEach Ship Do
    If Fleet$=fid Then CC$AddToSim False
  Next
EndSub

% DEL on fleet screen
Sub CC$FleetDeleteMember
  Local UI.Result
  If Fleet$ = Id Then
    % This is the fleet leader, so we'd better ask
    UI.Message Translate("Remove leader from fleet?"), Translate("Fleet"), Translate("Yes No")
    If UI.Result<>1 Then Return
  EndIf
  Fleet$ := 0
EndSub

% C-DEL on fleet screen
Sub CC$FleetDeleteFleet
  Local UI.Result
  Local fid = Fleet$
  UI.Message Translate("Dissolve this fleet?"), Translate("Fleet"), Translate("Yes No")
  If UI.Result=1 Then
    ForEach Ship Do If Id<>fid And Fleet$=fid Then Fleet$:=0
    Ship(fid).Fleet$:=0
  EndIf
EndSub

% 'b' in fleet screen
Sub CC$ChangeLeader
  Local UI.Result
  If Fleet$ = Id Then Return      % ship is already the leader
  UI.Message Translate("Make this ship the leader of the fleet?"), Translate("Fleet"), Translate("Yes No")
  If UI.Result=1 Then ChangeFleetLeader Fleet$, Id
EndSub

% S-F10 on ship screen
Sub CC$ShipGoToFleet
  Local UI.Result
  If Fleet$ Then
    UI.GotoScreen 10, Fleet$
  Else
    If Not FindShip(Fleet$) Then
      UI.Message Translate("You have not yet set up a fleet to join. Use the \"Start New Fleet\" command on the [F10] menu to start one."), Translate("Join Fleet"), Translate("OK")
    Else
      UI.ListFleets Loc.X, Loc.Y, "a", Translate("OK"), Translate("Join Fleet")
      If UI.Result Then Fleet$ := UI.Result
    EndIf
  EndIf
EndSub

EndIf % System.GUI

%%% Initialisation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Properties.
CreateShipProperty   Comment
CreatePlanetProperty Comment

% Keymaps. Note that those differ from PCC 1.x because we support multiple inheritance.
CreateKeymap Global, Ship, Planet, Base, Fleet

CreateKeymap ControlScreen(Global)
CreateKeymap ShipScreen(Ship, ControlScreen)
CreateKeymap PlanetScreen(Planet, ControlScreen)
CreateKeymap BaseScreen(Base, ControlScreen)
CreateKeymap FleetScreen(Fleet, ControlScreen)
CreateKeymap HistoryScreen(Ship, ControlScreen)
CreateKeymap AutoTaskScreen(ControlScreen)
CreateKeymap ShipTaskScreen(AutoTaskScreen)
CreateKeymap PlanetTaskScreen(AutoTaskScreen)
CreateKeymap BaseTaskScreen(AutoTaskScreen)

CreateKeymap Starchart(Global)
CreateKeymap ShipLock(Ship, Starchart)
CreateKeymap PlanetLock(Planet, Starchart)
CreateKeymap UnknownPlanetLock(PlanetLock)
CreateKeymap BaseLock(Base, Starchart)

CreateKeymap RaceScreen(Global)
CreateKeymap ShipBuildScreen(Global)

% Global Variables (System.Err is referenced by the interpreter core and therefore defined there).
Dim Shared UI.Result
Dim Shared UI.Directory
Dim Shared UI.Prefix
Dim Shared Cargo.Remainder
Dim Shared Build.Remainder

% Keybindings
Bind Base            "a"      := "CC$BuildAmmo"
Bind Base            "b"      := "CC$BuildShip"
Bind Base            "c-p"    := "CC$PickupBuild", "c-b" := "CC$PerformBuild"
Bind Base            "d"      := "CC$BuildStructures 2"
Bind Base            "f"      := "CC$ChangeFCode"
Bind Base            "m"      := "CC$ChangeMission"
Bind Base            "t"      := "CC$ChangeTech"
Bind Base            "c-c"    := "CC$GotoObject 1, FindShipCloningAt(Id)"
Bind Base            "ins"    := "Try CC$AddToSim"
Bind Base            "u"      := "CC$TransferPlanet 1"
Bind Base            "c"      := "CC$TransferPlanet 0"
Bind Base            "c-s"    := "CC$SearchBuild"             % FIXME: normally, this is on ShipBuildScreen only
Bind Base            "f9"     := "CC$SetPlanetComment"
Bind BaseScreen      "x"      := "CC$BaseXchg"
Bind BaseScreen      "f8"     := "UI.GotoScreen 2, Id"
Bind BaseScreen      "c-r"    := "CC$GotoObject 1, Shipyard.Id"
Bind BaseScreen      "tab"    := "CC$BuildShip"
Bind BaseScreen      "s"      := "CC$ShipCostCalc"
Bind BaseScreen      "h"      := "UI.Help 'pcc2:basescreen'"
Bind BaseScreen      "alt-h"  := "UI.Help 'pcc2:basescreen'"
Bind BaseScreen      "ret"    := "CC$GotoObject 13, Id"
Bind BaseTaskScreen  "ret"    := "CC$GotoObject 3, Id"
Bind BaseTaskScreen  "f8"     := "UI.GotoScreen 2, Id"
Bind BaseTaskScreen  "f9"     := "CC$SetPlanetComment"
Bind BaseTaskScreen  "h"      := "UI.Help 'pcc2:basetaskscreen'"
Bind BaseTaskScreen  "alt-h"  := "UI.Help 'pcc2:basetaskscreen'"

Bind Planet          "c-s"    := "CC$BuySupplies"
Bind Planet          "f"      := "CC$ChangeFCode"
Bind Planet          "f9"     := "CC$SetPlanetComment"
Bind Planet          "g"      := "CC$Give"
Bind Planet          "s"      := "CC$SellSupplies"
Bind Planet          "t"      := "CC$ChangeTaxes"
Bind Planet          "ins"    := "Try CC$AddToSim"
Bind Planet          "f8"     := "CC$BuildOrGotoBase"
Bind Planet          "u"      := "Try CC$TransferPlanet 1"
Bind Planet          "c"      := "Try CC$TransferPlanet 0"
Bind PlanetScreen    "b"      := "CC$BuildStructures 0"
Bind PlanetScreen    "tab"    := "CC$BuildStructures 0"
Bind PlanetScreen    "d"      := "CC$BuildStructures 2"
Bind PlanetScreen    "m"      := "CC$BuildStructures 1"
Bind PlanetScreen    "x"      := "CC$PlanetXchg"
Bind PlanetScreen    "h"      := "UI.Help 'pcc2:planetscreen'"
Bind PlanetScreen    "alt-h"  := "UI.Help 'pcc2:planetscreen'"
Bind PlanetScreen    "ret"    := "CC$GotoObject 12, Id"
Bind PlanetTaskScreen "ret"   := "CC$GotoObject 2, Id"
Bind PlanetTaskScreen "f8"    := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind PlanetTaskScreen "f9"    := "CC$SetPlanetComment"
Bind PlanetTaskScreen "h"     := "UI.Help 'pcc2:planettaskscreen'"
Bind PlanetTaskScreen "alt-h" := "UI.Help 'pcc2:planettaskscreen'"

Bind Ship            "c-f"    := "Try SetFCode Planet(Orbit$).FCode"
Bind Ship            "c-z"    := "CC$SetWaypoint Loc.X, Loc.Y"
Bind Ship            "c"      := "CC$ShipCargo"
Bind Ship            "c-j"    := "Try CC$TransferShip 2, 0"
Bind Ship            "c-p"    := "CC$ReviewShipTransfer 2, Transfer.Unload.Id"
Bind Ship            "c-s"    := "CC$ReviewShipTransfer 1, Transfer.Ship.Id"
Bind Ship            "e"      := "CC$ChangePE"
Bind Ship            "f"      := "CC$ChangeFCode"
Bind Ship            "f9"     := "CC$SetShipComment"
Bind Ship            "g"      := "CC$Give"
Bind Ship            "m"      := "CC$ChangeMission"
Bind Ship            "n"      := "CC$ShipName"
Bind Ship            "r"      := "Try CC$RemoteControl"
Bind Ship            "s"      := "CC$ShipSpec"
Bind Ship            "u"      := "CC$ShipUnload"
Bind Ship            "c-i"    := "Try UI.GotoScreen 1, Mission.Intercept"
Bind Ship            "c-t"    := "Try UI.GotoScreen 1, Mission.Tow"
Bind Ship            "a-i"    := "UI.Search 'Mission.Intercept=' & Id, 's2'"
Bind Ship            "a-t"    := "UI.Search 'Mission.Tow=' & Id, 's2'"
Bind Ship            "ins"    := "Try CC$AddToSim"
Bind Ship            "c-z"    := "CC$WithShipWaypoint 'SetWaypoint Loc.X, Loc.Y', True"
Bind Ship            "s-f10"  := "CC$ShipGoToFleet"
Bind ShipScreen      "f10"    := 0                        % cancel global binding!
Bind ShipScreen      "s-f1"   := "CC$ListShips Chart.X, Chart.Y, 'e' & Id"
Bind ShipScreen      "c-f1"   := "CC$ListShips UI.X, UI.Y, 'e' & Id"
Bind ShipScreen      "x"      := "CC$ShipXchg"
Bind ShipScreen      "f8"     := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind ShipScreen      "h"      := "UI.Help 'pcc2:shipscreen'"
Bind ShipScreen      "alt-h"  := "UI.Help 'pcc2:shipscreen'"
Bind ShipScreen      "f6"     := "CC$GotoObject 6, Id"
Bind ShipScreen      "ret"    := "CC$GotoObject 11, Id"
Bind ShipScreen      "d"      := "CC$TransferMulti"
Bind ShipScreen      "w"      := "CC$WithShipWaypoint 'CC$ChangeSpeed', False"
Bind ShipScreen      "a"      := "CC$WithShipWaypoint 'CC$ChangeWaypoint', False"
Bind ShipScreen      "c-w"    := "CC$WithShipWaypoint 'SetWaypoint UI.X, UI.Y', True"
Bind ShipScreen      "dblclick" := "CC$WithShipWaypoint 'SetWaypoint UI.X, UI.Y', True"
Bind ShipTaskScreen  "ret"    := "CC$GotoObject 1, Id"
Bind ShipTaskScreen  "w"      := "CC$WithShipWaypoint 'CC$ChangeSpeed', False, True"
Bind ShipTaskScreen  "c"      := "CC$ShipCargo"
Bind ShipTaskScreen  "f6"     := "CC$GotoObject 6, Id"
Bind ShipTaskScreen  "f9"     := "CC$SetShipComment"
Bind ShipTaskScreen  "h"      := "UI.Help 'pcc2:shiptaskscreen'"
Bind ShipTaskScreen  "alt-h"  := "UI.Help 'pcc2:shiptaskscreen'"
Bind ShipTaskScreen  "f8"     := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind HistoryScreen   "f1"     := "CC$ShipScreenFromHistory"
Bind HistoryScreen   "f6"     := "CC$GotoScreen 6"
Bind HistoryScreen   "h"      := "UI.Help 'pcc2:historyscreen'"
Bind HistoryScreen   "alt-h"  := "UI.Help 'pcc2:historyscreen'"
Bind HistoryScreen   "x"      := "CC$ShipXchg"

Bind ControlScreen   "."      := "CC$ToggleSelection"
Bind ControlScreen   "pgup"   := "CC$CSBrowse 0,0", "-" := "CC$CSBrowse 0,0", "wheelup" := "CC$CSBrowse 0,0"
Bind ControlScreen   "pgdn"   := "CC$CSBrowse 1,0", "+" := "CC$CSBrowse 1,0", "wheeldn" := "CC$CSBrowse 1,0"
Bind ControlScreen   "c-pgup" := "CC$CSBrowse 0,1", "c--" := "CC$CSBrowse 0,1", "c-wheelup" := "CC$CSBrowse 0,1"
Bind ControlScreen   "c-pgdn" := "CC$CSBrowse 1,1", "c-+" := "CC$CSBrowse 1,1", "c-wheeldn" := "CC$CSBrowse 1,1"
Bind ControlScreen   "s-f9"   := "CC$SetPlanetCommentAt Loc.X, Loc.Y"
Bind ControlScreen   "c-f9"   := "CC$SetPlanetCommentAt UI.X, UI.Y"
Bind ControlScreen   "s-f1"   := "CC$ListShips Chart.X, Chart.Y, 'e'"
Bind ControlScreen   "c-f1"   := "CC$ListShips UI.X, UI.Y, 'e'"
Bind ControlScreen   "s-f2"   := "CC$PlanetAt 2, Chart.X, Chart.Y"
Bind ControlScreen   "c-f2"   := "CC$PlanetAt 2, UI.X, UI.Y"
Bind ControlScreen   "s-f3"   := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind ControlScreen   "c-f3"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind ControlScreen   "s-f8"   := "CC$PlanetAt 3, Chart.X, Chart.Y"
Bind ControlScreen   "c-f8"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind ControlScreen   "s-f4"   := "CC$GotoChart Chart.X, Chart.Y"
Bind ControlScreen   "c-f4"   := "CC$GotoChart UI.X, UI.Y"
Bind ControlScreen   "esc"    := "UI.GotoScreen 0"
Bind ControlScreen   "s-l"    := "CC$ListShips Chart.X, Chart.Y, 'a'"
Bind ControlScreen   "c-l"    := "CC$ListShips UI.X, UI.Y, 'a'"
Bind ControlScreen   "l"      := "CC$ListShips UI.X, UI.Y, 'a'"
Bind ControlScreen   "f1"     := "CC$GotoScreen 1"
Bind ControlScreen   "f2"     := "CC$GotoScreen 2"
Bind ControlScreen   "f3"     := "CC$GotoScreen 3"
Bind ControlScreen   "f4"     := "CC$GotoChart Chart.X, Chart.Y"
Bind ControlScreen   "f5"     := "CC$PlanetInfo Chart.X, Chart.Y"
Bind ControlScreen   "s-f5"   := "CC$PlanetInfo Chart.X, Chart.Y"
Bind ControlScreen   "c-f5"   := "CC$PlanetInfo UI.X, UI.Y"
Bind ControlScreen   "f6"     := "CC$GotoScreen 6"
Bind ControlScreen   "s-f6"   := "CC$GotoScreen 6"
% Bind ControlScreen   "c-f6"   := "CC$GotoScreen 6"  -- how to do this?
Bind ControlScreen   "f7"     := "UI.Search"
Bind ControlScreen   "f10"    := "CC$GotoScreen 10"
Bind ControlScreen   "c-h"    := "UI.Search Chart.X & ', ' & Chart.Y, 'spbuo4'"

Bind FleetScreen     "b"      := "CC$ChangeLeader"
Bind FleetScreen     "c"      := "CC$ShipCargo"
Bind FleetScreen     "d"      := "CC$TransferMulti 1"
Bind FleetScreen     "f"      := "CC$ChangeFCode"
Bind FleetScreen     "g"      := "CC$Give"
Bind FleetScreen     "j"      := "CC$FleetJoin"
Bind FleetScreen     "n"      := "CC$FleetName"
Bind FleetScreen     "p"      := "CC$FleetSplit"
Bind FleetScreen     "r"      := "Try CC$RemoteControl"
Bind FleetScreen     "s"      := "CC$ShipSpec"
Bind FleetScreen     "t"      := "CC$TowFleetMember"
Bind FleetScreen     "u"      := "CC$ShipUnload"
Bind FleetScreen     "x"      := "CC$ShipXchg"
Bind FleetScreen     "f1"     := "CC$GotoObject 1, Id"
Bind FleetScreen     "f6"     := "CC$GotoObject 6, Id"
Bind FleetScreen     "f9"     := "CC$SetShipComment"
Bind FleetScreen     "ins"    := "Try CC$AddToSim"
Bind FleetScreen     "c-ins"  := "CC$FleetAddToSim"
Bind FleetScreen     "del"    := "CC$FleetDeleteMember"
Bind FleetScreen     "c-del"  := "CC$FleetDeleteFleet"
Bind FleetScreen     "c-w"    := "CC$WithFleetWaypoint 'SetWaypoint UI.X, UI.Y', True"
Bind FleetScreen     "w"      := "CC$WithFleetWaypoint 'CC$ChangeSpeed', False"
Bind FleetScreen     "a"      := "CC$WithFleetWaypoint 'CC$ChangeWaypoint', False"
Bind FleetScreen     "dblclick" := "CC$WithFleetWaypoint 'SetWaypoint UI.X, UI.Y', True"
Bind FleetScreen     "h"      := "UI.Help 'pcc2:fleetscreen'"
Bind FleetScreen     "alt-h"  := "UI.Help 'pcc2:fleetscreen'"

Bind ShipLock        "."      := "CC$ToggleSelection"
Bind ShipLock        "f1"     := "CC$GotoObject 1, Id"
Bind ShipLock        "s-f1"   := "CC$GotoObject 1, Id"
Bind ShipLock        "c-f1"   := "CC$GotoObject 1, Id"
Bind ShipLock        "f6"     := "CC$GotoObject 6, Id"
Bind ShipLock        "s-f6"   := "CC$GotoObject 6, Id"
Bind ShipLock        "c-f6"   := "CC$GotoObject 6, Id"
Bind ShipLock        "esc"    := "CC$HidePanel"
Bind PlanetLock      "."      := "CC$ToggleSelection"
Bind PlanetLock      "b"      := "CC$BaseLock"
Bind PlanetLock      "a"      := "Try CC$BuildStructures 0"
Bind PlanetLock      "esc"    := "CC$HidePanel"
Bind BaseLock        "p"      := "CC$PlanetLock"
Bind BaseLock        "esc"    := "CC$HidePanel"
Bind Starchart       "s-f9"   := "CC$SetPlanetCommentAt UI.X, UI.Y"
Bind Starchart       "c-f9"   := "CC$SetPlanetCommentAt UI.X, UI.Y"
Bind Starchart       "f7"     := "UI.Search"
Bind Starchart       "c-h"    := "UI.Search UI.X & ', ' & UI.Y, 'spbuo4'"
Bind Starchart       "esc"    := "UI.GotoScreen 0"
Bind Starchart       "f4"     := "UI.GotoScreen 0"
Bind Starchart       "f1"     := "CC$ListShips UI.X, UI.Y, 'e'"
Bind Starchart       "s-f1"   := "CC$ListShips UI.X, UI.Y, 'e'"
Bind Starchart       "c-f1"   := "CC$ListShips UI.X, UI.Y, 'e'"
Bind Starchart       "f2"     := "CC$PlanetAt 2, UI.X, UI.Y"
Bind Starchart       "s-f2"   := "CC$PlanetAt 2, UI.X, UI.Y"
Bind Starchart       "c-f2"   := "CC$PlanetAt 2, UI.X, UI.Y"
Bind Starchart       "f3"     := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "s-f3"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "c-f3"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "f6"     := "CC$GotoScreen 6"
Bind Starchart       "s-f6"   := "CC$GotoScreen 6"
Bind Starchart       "c-f6"   := "CC$GotoScreen 6"
Bind Starchart       "f8"     := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "s-f8"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "c-f8"   := "CC$PlanetAt 3, UI.X, UI.Y"
Bind Starchart       "l"      := "CC$ListShips UI.X, UI.Y, 'a'"
Bind Starchart       "f5"     := "CC$PlanetInfo UI.X, UI.Y"
Bind Starchart       "s-f5"   := "CC$PlanetInfo UI.X, UI.Y"
Bind Starchart       "c-f5"   := "CC$PlanetInfo UI.X, UI.Y"

Bind Starchart       "left"   := "CC$ScannerMoveX -10"
Bind Starchart       "right"  := "CC$ScannerMoveX 10"
Bind Starchart       "up"     := "CC$ScannerMoveY 10"
Bind Starchart       "down"   := "CC$ScannerMoveY -10"
Bind Starchart       "s-left" := "CC$ScannerMoveX -1"
Bind Starchart       "s-right":= "CC$ScannerMoveX 1"
Bind Starchart       "s-up"   := "CC$ScannerMoveY 1"
Bind Starchart       "s-down" := "CC$ScannerMoveY -1"
Bind Starchart       "c-left" := "CC$ScannerMoveX -100"
Bind Starchart       "c-right":= "CC$ScannerMoveX 100"
Bind Starchart       "c-up"   := "CC$ScannerMoveY 100"
Bind Starchart       "c-down" := "CC$ScannerMoveY -100"
Bind Starchart       "h"      := "UI.Help 'pcc2:starchart'"
Bind Starchart       "alt-h"  := "UI.Help 'pcc2:starchart'"

Bind RaceScreen      "f1"     := "CC$GotoScreen 1"
Bind RaceScreen      "f2"     := "CC$GotoScreen 2"
Bind RaceScreen      "f3"     := "CC$GotoScreen 3"
Bind RaceScreen      "f4"     := "UI.GotoScreen 4"
Bind RaceScreen      "f6"     := "CC$GotoScreen 6"
Bind RaceScreen      "f7"     := "UI.Search"
Bind RaceScreen      "f10"    := "CC$GotoScreen 10"
Bind RaceScreen      "m"      := "CC$ViewInbox"
Bind RaceScreen      "w"      := "CC$SendMessage"
Bind RaceScreen      "c-s"    := "SaveGame"
Bind RaceScreen      "esc"    := "CC$ExitRace"
Bind RaceScreen      "h"      := "UI.Help 'pcc2:racescreen'"
Bind RaceScreen      "alt-h"  := "UI.Help 'pcc2:racescreen'"
Bind RaceScreen      "ctrl-o" := "CC$Settings"

Bind Global          "a-c"    := "UI.PopupConsole"
Bind Global          "a-k"    := "UI.KeymapInfo"
Bind Global          "a-left" := "CC$SwitchSelection -1"
Bind Global          "a-right" := "CC$SwitchSelection 1"
Bind Global          "a-."    := "UI.SelectionManager"
Bind Global          "quit"   := "CC$ExitRace"
Bind Global          "bs"     := "CC$PopScreenHistory"
Bind Global          "a-bs"   := "CC$ListScreenHistory"

On EnterDirectory Do TryLoad "autoexec.q"

If System.GUI Then Print "[", System.Program, " ", System.Version, ", core.q ", CC$LibraryVersion, "]"
